# coding: utf-8

""" Miscellaneous astronomical velocity transformations. """

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Third-party
import numpy as np
from numpy import cos, sin

import astropy.coordinates as coord
import astropy.units as u
from astropy.coordinates.angles import rotation_matrix
from astropy.coordinates.builtin_frames.galactocentric import _ROLL0 as ROLL0

from .propermotion import pm_gal_to_icrs, pm_icrs_to_gal

__all__ = ["vgsr_to_vhel", "vhel_to_vgsr", "vgal_to_hel", "vhel_to_gal"]

# This is the default circular velocity and LSR peculiar velocity of the Sun
# TODO: make this a config item?
VCIRC = 220.*u.km/u.s
VLSR = [10., 5.25, 7.17]*u.km/u.s

def vgsr_to_vhel(coordinate, vgsr, vcirc=VCIRC, vlsr=VLSR):
    """
    Convert a radial velocity in the Galactic standard of rest (GSR) to
    a barycentric radial velocity.

    Parameters
    ----------
    coordinate : :class:`~astropy.coordinates.SkyCoord`
        An Astropy SkyCoord object or anything object that can be passed
        to the SkyCoord initializer.
    vgsr : :class:`~astropy.units.Quantity`
        GSR line-of-sight velocity.
    vcirc : :class:`~astropy.units.Quantity`
        Circular velocity of the Sun.
    vlsr : :class:`~astropy.units.Quantity`
        Velocity of the Sun relative to the local standard
        of rest (LSR).

    Returns
    -------
    vhel : :class:`~astropy.units.Quantity`
        Radial velocity in a barycentric rest frame.

    """

    c = coord.SkyCoord(coordinate)
    g = c.galactic
    l,b = g.l, g.b

    if not isinstance(vgsr, u.Quantity):
        raise TypeError("vgsr must be a Quantity subclass")

    # compute the velocity relative to the LSR
    lsr = vgsr - vcirc*sin(l)*cos(b)

    # velocity correction for Sun relative to LSR
    v_correct = vlsr[0]*cos(b)*cos(l) + \
        vlsr[1]*cos(b)*sin(l) + \
        vlsr[2]*sin(b)
    vhel = lsr - v_correct

    return vhel

def vhel_to_vgsr(coordinate, vhel, vcirc=VCIRC, vlsr=VLSR):
    """
    Convert a velocity from a heliocentric radial velocity to
    the Galactic standard of rest (GSR).

    Parameters
    ----------
    coordinate : :class:`~astropy.coordinates.SkyCoord`
        An Astropy SkyCoord object or anything object that can be passed
        to the SkyCoord initializer.
    vhel : :class:`~astropy.units.Quantity`
        Barycentric line-of-sight velocity.
    vcirc : :class:`~astropy.units.Quantity`
        Circular velocity of the Sun.
    vlsr : :class:`~astropy.units.Quantity`
        Velocity of the Sun relative to the local standard
        of rest (LSR).

    Returns
    -------
    vgsr : :class:`~astropy.units.Quantity`
        Radial velocity in a galactocentric rest frame.

    """

    c = coord.SkyCoord(coordinate)
    g = c.galactic
    l,b = g.l, g.b

    if not isinstance(vhel, u.Quantity):
        raise TypeError("vhel must be a Quantity subclass")

    lsr = vhel + vcirc*sin(l)*cos(b)

    # velocity correction for Sun relative to LSR
    v_correct = vlsr[0]*cos(b)*cos(l) + \
        vlsr[1]*cos(b)*sin(l) + \
        vlsr[2]*sin(b)
    vgsr = lsr + v_correct

    return vgsr

def _icrs_gctc_velocity_matrix(galactocentric_frame):
    """
    Construct a transformation matrix to go from heliocentric ICRS to a galactocentric
    frame. This is just a rotation and tilt which makes it approximately the same
    as transforming to Galactic coordinates. This only works for velocity because there
    is no shift due to the position of the Sun.
    """

    # define rotation matrix to align x(ICRS) with the vector to the Galactic center
    M1 = rotation_matrix(-galactocentric_frame.galcen_dec, 'y')
    M2 = rotation_matrix(galactocentric_frame.galcen_ra, 'z')

    # extra roll away from the Galactic x-z plane
    M3 = rotation_matrix(ROLL0 - galactocentric_frame.roll, 'x')

    # rotate about y' to account for tilt due to Sun's height above the plane
    z_d = (galactocentric_frame.z_sun / galactocentric_frame.galcen_distance).decompose()
    M4 = rotation_matrix(-np.arcsin(z_d), 'y')

    return M4*M3*M1*M2  # this is right: 4,3,1,2

def vgal_to_hel(coordinate, vxyz, vcirc=VCIRC, vlsr=VLSR, galactocentric_frame=None):
    r"""
    Convert a Galactocentric, cartesian velocity to a Heliocentric velocity in
    spherical coordinates (e.g., proper motion and radial velocity).

    The frame of the input coordinate determines the output frame of the proper motions.
    For example, if the input coordinate is in the ICRS frame, the proper motions
    returned will be  :math:`(\mu_\alpha\cos\delta,\mu_\delta)`. This function also
    handles array inputs (see examples below).

    Parameters
    ----------
    coordinate : :class:`~astropy.coordinates.SkyCoord`, :class:`~astropy.coordinates.BaseCoordinateFrame`
        This is most commonly a :class:`~astropy.coordinates.SkyCoord` object, but
        alternatively, it can be any coordinate frame object that is transformable to the
        Galactocentric frame.
    vxyz : :class:`~astropy.units.Quantity`, iterable
        Cartesian velocity components :math:`(v_x,v_y,v_z)`. This should either be a single
        :class:`~astropy.units.Quantity` object with shape (3,N), or an iterable
        object with 3 :class:`~astropy.units.Quantity` objects as elements.
    vcirc : :class:`~astropy.units.Quantity` (optional)
        Circular velocity of the Sun.
    vlsr : :class:`~astropy.units.Quantity` (optional)
        Velocity of the Sun relative to the local standard
        of rest (LSR).
    galactocentric_frame : :class:`~astropy.coordinates.Galactocentric` (optional)
        An instantiated :class:`~astropy.coordinates.Galactocentric` frame object with
        custom parameters for the Galactocentric coordinates. For example, if you want
        to set your own position of the Galactic center, you can pass in a frame with
        custom `galcen_ra` and `galcen_dec`.

    Returns
    -------
    pmv : tuple
        A tuple containing the proper motions (in Galactic coordinates) and
        radial velocity, all as :class:`~astropy.units.Quantity` objects.

    Examples
    --------

        >>> import astropy.units as u
        >>> import astropy.coordinates as coord
        >>> c = coord.Galactocentric(x=15.*u.kpc, y=13.*u.kpc, z=2.*u.kpc)
        >>> vxyz = [-115., 100., 95.]*u.km/u.s
        >>> icrs = c.transform_to(coord.ICRS)
        >>> vgal_to_hel(icrs, vxyz) # doctest: +FLOAT_CMP
        (<Quantity -0.876885123328934 mas / yr>, <Quantity 0.024501209459030334 mas / yr>, <Quantity -163.24449462243052 km / s>)

        >>> c = coord.Galactocentric([[15.,11.],[13,21.],[2.,-7]]*u.kpc)
        >>> vxyz = [[-115.,11.], [100.,-21.], [95.,103]]*u.km/u.s
        >>> icrs = c.transform_to(coord.ICRS)
        >>> vgal_to_hel(icrs, vxyz) # doctest: +FLOAT_CMP
        (<Quantity [-0.87688512,-0.91157482] mas / yr>, <Quantity [ 0.02450121,-0.86124895] mas / yr>, <Quantity [-163.24449462,-198.31241148] km / s>)

    """

    if galactocentric_frame is None:
        galactocentric_frame = coord.Galactocentric

    # so I don't accidentally modify in place
    vxyz = vxyz.copy()

    c = coordinate
    R = _icrs_gctc_velocity_matrix(galactocentric_frame)

    # remove circular and LSR velocities
    vxyz[1] = vxyz[1] - vcirc
    for i in range(3):
        vxyz[i] = vxyz[i] - vlsr[i]

    orig_shape = vxyz.shape
    # v_icrs = np.linalg.inv(R).dot(vxyz.reshape(vxyz.shape[0], np.prod(vxyz.shape[1:]))).reshape(orig_shape)
    v_icrs = R.T.dot(vxyz.reshape(vxyz.shape[0], np.prod(vxyz.shape[1:]))).reshape(orig_shape)

    # get cartesian heliocentric
    x_icrs = c.transform_to(coord.ICRS).cartesian.xyz
    d = np.sqrt(np.sum(x_icrs**2, axis=0))
    dxy = np.sqrt(x_icrs[0]**2 + x_icrs[1]**2)

    vr = np.sum(x_icrs * v_icrs, axis=0) / d

    mua = ((x_icrs[0]*v_icrs[1] - v_icrs[0]*x_icrs[1]) / dxy**2).to(u.mas/u.yr, equivalencies=u.dimensionless_angles())
    mua_cosd = (mua * dxy / d).to(u.mas/u.yr, equivalencies=u.dimensionless_angles())
    mud = (-(x_icrs[2]*(x_icrs[0]*v_icrs[0] + x_icrs[1]*v_icrs[1]) - dxy**2*v_icrs[2]) / d**2 / dxy).to(u.mas/u.yr, equivalencies=u.dimensionless_angles())

    pm_radec = (mua_cosd, mud)

    if c.name == 'icrs':
        pm = u.Quantity(map(np.atleast_1d,pm_radec))

    elif c.name == 'galactic':
        # transform to ICRS proper motions
        pm = pm_icrs_to_gal(c, pm_radec)

    else:
        raise NotImplementedError("Proper motions in the {} system are not "
                                  "currently supported.".format(c.name))

    if c.isscalar:
        vr = vr.reshape(())
        pm = (pm[0].reshape(()), pm[1].reshape(()))

    return tuple(pm) + (vr,)

def vhel_to_gal(coordinate, pm, rv, vcirc=VCIRC, vlsr=VLSR, galactocentric_frame=None):
    r"""
    Convert a Heliocentric velocity in spherical coordinates (e.g., proper motion
    and radial velocity) in the ICRS or Galactic frame to a Galactocentric, cartesian
    velocity.

    The frame of the input coordinate determines how to interpret the given
    proper motions. For example, if the input coordinate is in the ICRS frame, the
    proper motions are assumed to be :math:`(\mu_\alpha\cos\delta,\mu_\delta)`. This
    function also handles array inputs (see examples below).

    TODO: Roundtrip using galactic coordinates only maintains relative precision
    of ~1E-5. Why?

    Parameters
    ----------
    coordinate : :class:`~astropy.coordinates.SkyCoord`, :class:`~astropy.coordinates.BaseCoordinateFrame`
        This is most commonly a :class:`~astropy.coordinates.SkyCoord` object, but
        alternatively, it can be any coordinate frame object that is transformable to the
        Galactocentric frame.
    pm : :class:`~astropy.units.Quantity` or iterable of :class:`~astropy.units.Quantity` objects
        Proper motion in the same frame as the coordinate. For example, if your input
        coordinate is in :class:`~astropy.coordinates.ICRS`, then the proper motion is
        assumed to be in this frame as well. The order of elements should always be
        proper motion in (longitude, latitude), and should have shape (2,N). The longitude
        component is assumed to have the cosine of the latitude already multiplied in, so
        that in ICRS, for example, this would be :math:`\mu_\alpha\cos\delta`.
    rv : :class:`~astropy.units.Quantity`
        Barycentric radial velocity. Should have shape (1,N) or (N,).
    vcirc : :class:`~astropy.units.Quantity` (optional)
        Circular velocity of the Sun.
    vlsr : :class:`~astropy.units.Quantity` (optional)
        Velocity of the Sun relative to the local standard
        of rest (LSR).
    galactocentric_frame : :class:`~astropy.coordinates.Galactocentric` (optional)
        An instantiated :class:`~astropy.coordinates.Galactocentric` frame object with
        custom parameters for the Galactocentric coordinates. For example, if you want
        to set your own position of the Galactic center, you can pass in a frame with
        custom `galcen_ra` and `galcen_dec`.

    Returns
    -------
    vxyz : :class:`~astropy.units.Quantity` (optional)
        Cartesian velocity components (U,V,W). A :class:`~astropy.units.Quantity`
        object with shape (3,N).

    Examples
    --------

        >>> import astropy.units as u
        >>> import astropy.coordinates as coord
        >>> c = coord.SkyCoord(ra=196.5*u.degree, dec=-10.33*u.deg, distance=16.2*u.kpc)
        >>> pm = [-1.53, 3.5]*u.mas/u.yr
        >>> rv = 161.4*u.km/u.s
        >>> vhel_to_gal(c, pm=pm, rv=rv) # doctest: +FLOAT_CMP
        <Quantity [-137.29984564, 262.64052249, 305.50786499] km / s>

        >>> c = coord.SkyCoord(ra=[196.5,51.3]*u.degree, dec=[-10.33,2.1]*u.deg, distance=[16.2,11.]*u.kpc)
        >>> pm = [[-1.53,4.5], [3.5,10.9]]*u.mas/u.yr
        >>> rv = [161.4,-210.2]*u.km/u.s
        >>> vhel_to_gal(c, pm=pm, rv=rv) # doctest: +FLOAT_CMP
        <Quantity [[-137.29984564,-212.10415701],
                   [ 262.64052249, 496.85687803],
                   [ 305.50786499, 554.16562628]] km / s>

    """

    if galactocentric_frame is None:
        galactocentric_frame = coord.Galactocentric

    # make sure this is a coordinate and get the frame for later use
    c = coordinate

    if c.name == 'icrs':
        pm_radec = u.Quantity(map(np.atleast_1d,pm))
        icrs = c

    elif c.name == 'galactic':
        # transform to ICRS proper motions
        pm_radec = pm_gal_to_icrs(c, pm)
        icrs = c.transform_to(coord.ICRS)

    else:
        raise NotImplementedError("Proper motions in the {} system are not "
                                  "currently supported.".format(c.name))

    # I'm so fired
    a,d,D = icrs.ra, icrs.dec, c.distance

    # proper motion components: longitude, latitude
    mura_cosdec, mudec = pm_radec
    vra = (D*mura_cosdec).to(rv.unit, equivalencies=u.dimensionless_angles())
    vdec = (D*mudec).to(rv.unit, equivalencies=u.dimensionless_angles())

    v_icrs = [rv*np.cos(a)*np.cos(d) - vra*np.sin(a) - vdec*np.cos(a)*np.sin(d),
              rv*np.sin(a)*np.cos(d) + vra*np.cos(a) - vdec*np.sin(a)*np.sin(d),
              rv*np.sin(d) + vdec*np.cos(d)]
    v_icrs = np.array([v.to(u.km/u.s).value for v in v_icrs]) * u.km/u.s

    R = _icrs_gctc_velocity_matrix(galactocentric_frame)

    orig_shape = v_icrs.shape
    v_gc = R.dot(v_icrs.reshape(v_icrs.shape[0], np.prod(v_icrs.shape[1:]))).reshape(orig_shape)

    # remove circular and LSR velocities
    v_gc[1] = v_gc[1] + vcirc
    for i in range(3):
        v_gc[i] = v_gc[i] + vlsr[i]

    if c.isscalar:
        return v_gc.reshape((3,))
    else:
        return v_gc



# coding: utf-8

""" test reading/writing potentials to files """

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Third-party
import astropy.units as u
from astropy.utils.data import get_pkg_data_filename
from astropy.tests.helper import quantity_allclose
import numpy as np

# Project
from ..io import load, save
from ..core import CompositePotential
from ..builtin import IsochronePotential, KeplerPotential, CCompositePotential
from ..builtin.special import LM10Potential
from ...units import DimensionlessUnitSystem, galactic

def test_read_plummer():
    potential = load(get_pkg_data_filename('Plummer.yml'))
    assert np.allclose(potential.parameters['m'].value, 100000000000.)
    assert np.allclose(potential.parameters['b'].value, 0.26)
    assert potential.parameters['b'].unit == u.kpc

def test_read_harmonic_oscillator():
    potential = load(get_pkg_data_filename('HarmonicOscillator1D.yml'))
    assert isinstance(potential.units, DimensionlessUnitSystem)

def test_read_composite():
    potential = load(get_pkg_data_filename('Composite.yml'))
    assert 'halo' in potential.keys()
    assert 'disk' in potential.keys()
    assert str(potential) == "CompositePotential"
    assert potential.units['length'] == u.kpc
    assert potential.units['speed'] == u.km/u.s

def test_read_lm10():
    potential = load(get_pkg_data_filename('lm10.yml'))
    assert 'halo' in potential.keys()
    assert 'disk' in potential.keys()
    assert str(potential) == "LM10Potential"
    assert np.allclose(potential['disk'].parameters['a'].value, 10)
    assert np.allclose(potential['disk'].parameters['b'].value, 0.26)
    assert np.allclose(potential['disk'].parameters['m'].value, 150000.)

def test_write_isochrone(tmpdir):
    tmp_filename = str(tmpdir.join("potential.yml"))

    # try a simple potential
    potential = IsochronePotential(m=1E11, b=0.76, units=galactic)

    with open(tmp_filename,'w') as f:
        save(potential, f)

    save(potential, tmp_filename)
    p = load(tmp_filename)

def test_write_isochrone_units(tmpdir):
    tmp_filename = str(tmpdir.join("potential.yml"))

    # try a simple potential with units
    potential = IsochronePotential(m=1E11*u.Msun, b=0.76*u.kpc, units=galactic)

    with open(tmp_filename,'w') as f:
        save(potential, f)

    save(potential, tmp_filename)
    p = load(tmp_filename)

def test_write_lm10(tmpdir):
    tmp_filename = str(tmpdir.join("potential.yml"))

    # more complex
    potential = LM10Potential(disk=dict(m=5E12*u.Msun))
    potential_default = LM10Potential()
    v1 = potential.value([4., 0, 0])
    v2 = potential_default.value([4., 0, 0])

    with open(tmp_filename,'w') as f:
        save(potential, f)

    save(potential, tmp_filename)
    p = load(tmp_filename)
    assert quantity_allclose(p['disk'].parameters['m'], 5E12*u.Msun)
    assert quantity_allclose(v1, p.value([4.,0,0]))
    assert not quantity_allclose(v2, p.value([4.,0,0]))

def test_write_composite(tmpdir):
    tmp_filename = str(tmpdir.join("potential.yml"))
    print(tmp_filename)

    # composite potential
    potential = CompositePotential(halo=KeplerPotential(m=1E11, units=galactic),
                                   bulge=IsochronePotential(m=1E11, b=0.76, units=galactic))
    save(potential, tmp_filename)
    p = load(tmp_filename)

def test_write_ccomposite(tmpdir):
    tmp_filename = str(tmpdir.join("potential.yml"))

    # composite potential
    potential = CCompositePotential(halo=KeplerPotential(m=1E11, units=galactic),
                                    bulge=IsochronePotential(m=1E11, b=0.76, units=galactic))
    save(potential, tmp_filename)
    p = load(tmp_filename)

def test_units(tmpdir):
    import astropy.units as u

    tmp_filename = str(tmpdir.join("potential.yml"))

    # try a simple potential
    potential = KeplerPotential(m=1E11, units=[u.kpc,u.Gyr,u.Msun,u.radian])
    save(potential, tmp_filename)
    p = load(tmp_filename)



from setuptools.command.build_py import build_py as SetuptoolsBuildPy

from ..utils import _get_platlib_dir


class AstropyBuildPy(SetuptoolsBuildPy):
    user_options = SetuptoolsBuildPy.user_options[:]
    boolean_options = SetuptoolsBuildPy.boolean_options[:]

    def finalize_options(self):
        # Update build_lib settings from the build command to always put
        # build files in platform-specific subdirectories of build/, even
        # for projects with only pure-Python source (this is desirable
        # specifically for support of multiple Python version).
        build_cmd = self.get_finalized_command('build')
        platlib_dir = _get_platlib_dir(build_cmd)

        build_cmd.build_purelib = platlib_dir
        build_cmd.build_lib = platlib_dir
        self.build_lib = platlib_dir

        SetuptoolsBuildPy.finalize_options(self)

    def run_2to3(self, files, doctests=False):
        # Filter the files to exclude things that shouldn't be 2to3'd
        skip_2to3 = self.distribution.skip_2to3
        filtered_files = []
        for filename in files:
            for package in skip_2to3:
                if filename[len(self.build_lib) + 1:].startswith(package):
                    break
            else:
                filtered_files.append(filename)

        SetuptoolsBuildPy.run_2to3(self, filtered_files, doctests)

    def run(self):
        # first run the normal build_py
        SetuptoolsBuildPy.run(self)



# coding: utf-8

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Third-party
import astropy.units as u
import matplotlib.pyplot as pl
import numpy as np
import pytest

# Custom
from ....potential import SphericalNFWPotential
from ....dynamics import CartesianPhaseSpacePosition
from ....integrate import DOPRI853Integrator
from ....units import galactic

# Project
from ..core import mock_stream, streakline_stream, fardal_stream, dissolved_fardal_stream

def test_mock_stream():
    potential = SphericalNFWPotential(v_c=0.2, r_s=20., units=galactic)

    w0 = CartesianPhaseSpacePosition(pos=[0.,15.,0]*u.kpc,
                                     vel=[-0.13,0,0]*u.kpc/u.Myr)
    prog = potential.integrate_orbit(w0, dt=-2., n_steps=1023)
    prog = prog[::-1]

    k_mean = [1.,0.,0.,0.,1.,0.]
    k_disp = [0.,0.,0.,0.,0.,0.]
    stream = mock_stream(potential, prog, k_mean=k_mean, k_disp=k_disp,
                         prog_mass=1E4, Integrator=DOPRI853Integrator)

    # fig = prog.plot(subplots_kwargs=dict(sharex=False,sharey=False))
    # fig = stream.plot(color='#ff0000', alpha=0.5, axes=fig.axes)
    # fig = stream.plot()
    # pl.show()

    assert stream.pos.shape == (3,2048) # two particles per step

    diff = np.abs(stream[-2:].pos - prog[-1].pos)
    assert np.allclose(diff[0].value, 0.)
    assert np.allclose(diff[1,0].value, diff[1,1].value)
    assert np.allclose(diff[2].value, 0.)

mock_funcs = [streakline_stream, fardal_stream, dissolved_fardal_stream]
all_extra_args = [dict(), dict(), dict(t_disrupt=-250.*u.Myr)]
@pytest.mark.parametrize("mock_func, extra_kwargs", zip(mock_funcs, all_extra_args))
def test_each_type(mock_func, extra_kwargs):
    potential = SphericalNFWPotential(v_c=0.2, r_s=20., units=galactic)

    w0 = CartesianPhaseSpacePosition(pos=[0.,15.,0]*u.kpc,
                                     vel=[-0.13,0,0]*u.kpc/u.Myr)
    prog = potential.integrate_orbit(w0, dt=-2., n_steps=1023)
    prog = prog[::-1]

    stream = mock_func(potential, prog_orbit=prog, prog_mass=1E4,
                       Integrator=DOPRI853Integrator, **extra_kwargs)

    # fig = prog.plot(subplots_kwargs=dict(sharex=False,sharey=False))
    # fig = stream.plot(color='#ff0000', alpha=0.5, axes=fig.axes)
    # fig = stream.plot()
    # pl.show()

    assert prog.t.shape == (1024,)
    assert stream.pos.shape == (3,2048) # two particles per step



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This sphinx extension builds off of `sphinx.ext.autosummary` to
clean up some issues it presents in the Astropy docs.

The main issue this fixes is the summary tables getting cut off before the
end of the sentence in some cases.

Note: Sphinx 1.2 appears to have fixed the the main issues in the stock
autosummary extension that are addressed by this extension.  So use of this
extension with newer versions of Sphinx is deprecated.
"""

import re

from distutils.version import LooseVersion

import sphinx

from sphinx.ext.autosummary import Autosummary

from ...utils import deprecated

# used in AstropyAutosummary.get_items
_itemsummrex = re.compile(r'^([A-Z].*?\.(?:\s|$))')


@deprecated('1.0', message='AstropyAutosummary is only needed when used '
                           'with Sphinx versions less than 1.2')
class AstropyAutosummary(Autosummary):
    def get_items(self, names):
        """Try to import the given names, and return a list of
        ``[(name, signature, summary_string, real_name), ...]``.
        """
        from sphinx.ext.autosummary import (get_import_prefixes_from_env,
                                            import_by_name, get_documenter,
                                            mangle_signature)

        env = self.state.document.settings.env

        prefixes = get_import_prefixes_from_env(env)

        items = []

        max_item_chars = 50

        for name in names:
            display_name = name
            if name.startswith('~'):
                name = name[1:]
                display_name = name.split('.')[-1]

            try:
                import_by_name_values = import_by_name(name, prefixes=prefixes)
            except ImportError:
                self.warn('[astropyautosummary] failed to import %s' % name)
                items.append((name, '', '', name))
                continue

            # to accommodate Sphinx v1.2.2 and v1.2.3
            if len(import_by_name_values) == 3:
                real_name, obj, parent = import_by_name_values
            elif len(import_by_name_values) == 4:
                real_name, obj, parent, module_name = import_by_name_values

            # NB. using real_name here is important, since Documenters
            #     handle module prefixes slightly differently
            documenter = get_documenter(obj, parent)(self, real_name)
            if not documenter.parse_name():
                self.warn('[astropyautosummary] failed to parse name %s' % real_name)
                items.append((display_name, '', '', real_name))
                continue
            if not documenter.import_object():
                self.warn('[astropyautosummary] failed to import object %s' % real_name)
                items.append((display_name, '', '', real_name))
                continue

            # -- Grab the signature

            sig = documenter.format_signature()
            if not sig:
                sig = ''
            else:
                max_chars = max(10, max_item_chars - len(display_name))
                sig = mangle_signature(sig, max_chars=max_chars)
                sig = sig.replace('*', r'\*')

            # -- Grab the summary

            doc = list(documenter.process_doc(documenter.get_doc()))

            while doc and not doc[0].strip():
                doc.pop(0)
            m = _itemsummrex.search(" ".join(doc).strip())
            if m:
                summary = m.group(1).strip()
            elif doc:
                summary = doc[0].strip()
            else:
                summary = ''

            items.append((display_name, sig, summary, real_name))

        return items


def setup(app):
    # need autosummary, of course
    app.setup_extension('sphinx.ext.autosummary')

    # Don't make the replacement if Sphinx is at least 1.2
    if LooseVersion(sphinx.__version__) < LooseVersion('1.2.0'):
        # this replaces the default autosummary with the astropy one
        app.add_directive('autosummary', AstropyAutosummary)
    elif LooseVersion(sphinx.__version__) < LooseVersion('1.3.2'):
        # Patch Autosummary again, but to work around an upstream bug; see
        # https://github.com/astropy/astropy-helpers/issues/172
        class PatchedAutosummary(Autosummary):
            def get_items(self, names):
                self.genopt['imported-members'] = True
                return Autosummary.get_items(self, names)

        app.add_directive('autosummary', PatchedAutosummary)



#################
#   Potentials  #
#################
import numpy as np
from scipy.integrate import quad, ode

Grav = 430091.5694

class LMPot(object):
    """ Potential used in Law-Majewski 2010 """
    def __init__(self):
        """ Best-fit parameters - units = kpc, km/s and 10^11 M_sol """
        self.M_disk = 1.
        self.a_disk = 6.5
        self.b_disk = 0.26
        self.M_bulge = 0.34
        self.c_bulge = 0.7
        self.vhalo2 = 121.7**2
        phi = 97./180.*np.pi
        q1 = 1.38
        q2 = 1.
        self.C_1 = (np.cos(phi)/q1)**2+(np.sin(phi)/q2)**2
        self.C_2 = (np.cos(phi)/q2)**2+(np.sin(phi)/q1)**2
        self.C_3 = 2.*np.sin(phi)*np.cos(phi)*(1./q1/q1-1./q2/q2)
        self.q_z = 1.36
        self.rhalo2 = 144.
        rot90 = np.array([[0.,1.],[-1.,0.]])
        self.rotmatrix =  np.dot(rot90,np.linalg.svd(np.array([[self.C_1,self.C_3/2.],[self.C_3/2.,self.C_2]]))[0])
        self.invrotmatrix = np.linalg.inv(self.rotmatrix)

    def disk_pot(self,x,y,z):
        R = np.sqrt(x*x+y*y)
        return -Grav*self.M_disk/np.sqrt(R*R+(self.a_disk+np.sqrt(z*z+self.b_disk*self.b_disk))**2)

    def disk_force(self,x,y,z):
        R = np.sqrt(x*x+y*y)
        e = self.a_disk+np.sqrt(z*z+self.b_disk*self.b_disk)
        d = -Grav*self.M_disk/np.sqrt(R*R+e**2)**3
        return np.array([x*d, y*d, z*d*e/(e-self.a_disk)])

    def bulge_pot(self,x,y,z):
        r = np.sqrt(x*x+y*y+z*z)
        return -Grav*self.M_bulge/(r+self.c_bulge)

    def bulge_force(self,x,y,z):
        r = np.sqrt(x*x+y*y+z*z)
        if(r==0.):
            return -Grav*self.M_bulge/(r+self.c_bulge)**2
        else:
            return -Grav*self.M_bulge/(r+self.c_bulge)**2/r*np.array([x,y,z])

    def halo_pot(self,x,y,z):
        return self.vhalo2*np.log(self.C_1*x*x+self.C_2*y*y+self.C_3*x*y+(z/self.q_z)**2+self.rhalo2)

    def halo_force(self,x,y,z):
        p = -self.vhalo2/(self.C_1*x*x+self.C_2*y*y+self.C_3*x*y+(z/self.q_z)**2+self.rhalo2)
        return np.array([(2.*x*self.C_1+self.C_3*y)*p,(2.*y*self.C_2+self.C_3*x)*p,2.*z*p/self.q_z**2])

    def tot_pot(self,x,y,z):
        return self.disk_pot(x,y,z)+self.bulge_pot(x,y,z)+self.halo_pot(x,y,z)

    def H(self,X):
        return 0.5*np.sum(X[3:]**2)+self.tot_pot(*X[:3])

    def tot_force(self,x,y,z):
        return self.disk_force(x,y,z)+self.bulge_force(x,y,z)+self.halo_force(x,y,z)

    def coordrot(self,x,y):
        return np.dot(self.rotmatrix,np.array([x,y]))

    def invcoordrot(self,x,y):
        return np.dot(self.invrotmatrix,np.array([x,y]))


class log_triax(object):
    """ test triaxial logarithmic potential
        Phi(x,y,z) = 0.5 v_c^2 log(Rc^2+x^2+(y/qy)^2+(z/qz)^2) """
    def __init__(self,vc,Rc,qy,qz):
        self.vc2=vc*vc
        self.Rc2=Rc*Rc
        self.qy2=qy*qy
        self.qz2=qz*qz

    def pot(self,x,y,z):
        return self.vc2/2.*np.log(self.Rc2+x*x+y*y/self.qy2+z*z/self.qz2)

    def H(self, X):
        return 0.5*np.sum(X[3:]**2)+self.pot(*X[:3])

    def tot_force(self,x,y,z):
        p = self.Rc2+x*x+y*y/self.qy2+z*z/self.qz2
        return -self.vc2/p*np.array([x,y/self.qy2,z/self.qz2])


class quartic(object):
    """ Quartic potential
        Phi(x,y,z) = 0.25(lam[0] x^4+lam[1] y^4+lam[2] z^4.
    """

    def __init__(self,lam = np.array([1.,0.8,3.3])):
        self.lambd = lam

    def H(self,x):
        """ Quartic potential Hamiltonian """
        return 0.5*np.sum(x[3:]**2+0.5*self.lambd*x[:3]**4)

    def tot_force(self,x,y,z):
        """ Derivatives of quartic potential for orbit integration """
        return np.array([-self.lambd[0]*x**3,-self.lambd[1]*y**3,-self.lambd[2]*z**3])

    def action(self,x):
        """ Find true action for quartic potential \Phi = \sum_i 0.25*x_i**4 """
        acts = np.ones(3)
        for i in range(3):
            En = 0.5*x[i+3]**2+0.25*self.lambd[i]*x[i]**4
            xlim=(4.*En/self.lambd[i])**0.25
            acts[i]=2.*quad(lambda y:np.sqrt(2.*En-0.5*self.lambd[i]*y**4),0.,xlim)[0]/np.pi
        return acts

    def freq(self,x):
        """ Find true freq. for quartic potential \Phi = 0.25*x**4 """
        freq = np.ones(3)
        for i in range(3):
            En = 0.5*x[i+3]**2+0.25*self.lambd[i]*x[i]**4
            xlim=(4.*En/self.lambd[i])**0.25
            freq[i]=np.pi/quad(lambda y:2./np.sqrt(2.*En-0.5*self.lambd[i]*y**4),0.,xlim)[0]
        return freq

import sys
# sys.path.append("new_struct")
# import triax_py

class stackel_triax(object):
    """ For interface with C code to find actions in triaxial Stackel potential """
    def __init__(self):
        pass

    def H(self,x):
        """ triaxial stackel potential Hamiltonian """
        return triax_py.Stack_Triax_H(x)

    def tot_force(self,x,y,z):
        """ Derivatives of triaxial stackel potential for orbit integration """
        X = np.array([x,y,z])
        return triax_py.Stack_Triax_Forces(X)

    def action(self,x):
        """ Find true action for triaxial stackel potential """
        return triax_py.Stack_Triax_Actions(x)

    def freq(self,x):
        """ Find true action for triaxial stackel potential """
        return triax_py.Stack_Triax_Freqs(x)


class harmonic_oscillator(object):
    """
        Triaxial harmonic oscillator
        Phi(x,y,z) = 0.5*(omega[0]^2 x^2+omega[1]^2 y^2 + omega[2]^2 z^2
    """
    def __init__(self,omega=np.array([1.,1.,1.])):
        self.omega = omega

    def H(self,x):
        """ Hamiltonian """
        return 0.5*np.sum(x[3:]**2+(self.omega*x[:3])**2)

    def tot_force(self,x,y,z):
        """ Derivatives of ho potential for orbit integration """
        return -np.array([self.omega[0]**2*x,self.omega[1]**2*y,self.omega[2]**2*z])


class isochrone(object):
    """
        Isochrone potential
        Phi(r) = -GM/(b+sqrt(b^2+r^2))
    """
    def __init__(self,par = np.array([1./Grav,4.2,0.])):
        """ params = {M, b, r0} """
        self.params = par

    def H(self,x):
        """ Hamiltonian """
        r = (np.sqrt(np.sum(x[:3]**2))-self.params[2])**2
        return 0.5*np.sum(x[3:]**2)-Grav*self.params[0]/(self.params[1]+np.sqrt(self.params[1]**2+r))

    def pot(self,x):
        r = (np.sqrt(np.sum(x[:3]**2))-self.params[2])**2
        return -Grav*self.params[0]/(self.params[1]+np.sqrt(self.params[1]**2+r))

    def tot_force(self,x,y,z):
        """ Derivatives of isochrone potential for orbit integration """
        r = (np.sqrt(x*x+y*y+z*z)-self.params[2])**2
        fac = np.sqrt(r)/(np.sqrt(r)+self.params[2])
        return np.array([x,y,z])*fac*-Grav*self.params[0]/(self.params[1]+np.sqrt(self.params[1]**2+r))**2/np.sqrt(self.params[1]**2+r)


def orbit_derivs(t,x,Pot):
    """ Simple interface for derivatives for orbit integration
        t = time
        x = Cartesian coordinates
        Pot is an object which has a function tot_force(x,y,z) which
        calculates the total force at Cartesian x,y,z """
    X=x[0]
    Y=x[1]
    Z=x[2]
    return np.concatenate((x[3:],Pot.tot_force(X,Y,Z)))

def orbit_derivs2(x,t,Pot):
    return orbit_derivs(t,x,Pot)

import warnings

def orbit_integrate(x,tmax,Pot):
    """ Integrates an orbit with initial coordinates x for time tmax in
    potential Pot using Dormund Prince 8 adaptive step size """
    solver = ode(orbit_derivs).set_integrator('dopri5', n_steps=1, rtol=1e-10,atol=1e-10)
    solver.set_initial_value(x,0.).set_f_params(Pot)
    solver._integrator.iwork[2] = -1
    warnings.filterwarnings("ignore", category=UserWarning)
    t = np.array([0.])
    while solver.t < tmax:
        solver.integrate(tmax)
        x=np.vstack((x,solver.y))
        t=np.append(t,solver.t)
    warnings.resetwarnings()
    return x,t

def leapfrog_integrator(x,tmax,NT,Pot):
    deltat = tmax/NT
    h = deltat/100.
    t = 0.
    counter = 0
    X = np.copy(x)
    results = np.array([x])
    while(t<tmax):
        X[3:] += 0.5*h*Pot.tot_force(X[0],X[1],X[2])
        X[:3] += h*X[3:]
        X[3:] += 0.5*h*Pot.tot_force(X[0],X[1],X[2])
        # if(t==0.1):
        if(counter % 100 == 0):
            results=np.vstack((results,X))
        t+=h
        counter+=1
    return results



# coding: utf-8
"""
    Test the special potentials...
"""

from __future__ import absolute_import, unicode_literals, division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Third-party
import astropy.units as u

# This project
from .helpers import CompositePotentialTestBase
from ..builtin.special import *

class TestLM10Potential(CompositePotentialTestBase):
    potential = LM10Potential()
    w0 = [8.,0.,0.,0.,0.22,0.1]

class TestLM10Potential2(CompositePotentialTestBase):
    potential = LM10Potential(disk={'m': 5E10*u.Msun}, bulge={'m': 5E10*u.Msun})
    w0 = [8.,0.,0.,0.,0.22,0.1]

# class TestTriaxialMW(PotentialTestBase):
#     potential = TriaxialMWPotential()
#     w0 = [8.,0.,0.,0.,0.22,0.1]



import sys

from textwrap import dedent

import pytest

from ..autodoc_enhancements import type_object_attrgetter


# Define test classes outside the class; otherwise there is flakiness with the
# details of how exec works on different Python versions
class Meta(type):
    @property
    def foo(cls):
        return 'foo'

if sys.version_info[0] < 3:
    exec(dedent("""
        class MyClass(object):
            __metaclass__ = Meta
            @property
            def foo(self):
                \"\"\"Docstring for MyClass.foo property.\"\"\"
                return 'myfoo'
    """))
else:
    exec(dedent("""
        class MyClass(metaclass=Meta):
            @property
            def foo(self):
                \"\"\"Docstring for MyClass.foo property.\"\"\"
                return 'myfoo'
    """))


def test_type_attrgetter():
    """
    This test essentially reproduces the docstring for
    `type_object_attrgetter`.

    Sphinx itself tests the custom attrgetter feature; see:
    https://bitbucket.org/birkenfeld/sphinx/src/40bd03003ac6fe274ccf3c80d7727509e00a69ea/tests/test_autodoc.py?at=default#cl-502
    so rather than a full end-to-end functional test it's simple enough to just
    test that this function does what it needs to do.
    """

    assert getattr(MyClass, 'foo') == 'foo'
    obj = type_object_attrgetter(MyClass, 'foo')
    assert isinstance(obj, property)
    assert obj.__doc__ == 'Docstring for MyClass.foo property.'

    with pytest.raises(AttributeError):
        type_object_attrgetter(MyClass, 'susy')

    assert type_object_attrgetter(MyClass, 'susy', 'default') == 'default'
    assert type_object_attrgetter(MyClass, '__dict__') == MyClass.__dict__



# coding: utf-8

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Standard lib
from collections import OrderedDict

# Third-party
import astropy.units as u
import numpy as np

# Project
# from .cpotential import CCompositePotential
# from ..core import CompositePotential
from .cybuiltin import HernquistPotential, MiyamotoNagaiPotential, \
    LeeSutoTriaxialNFWPotential, SphericalNFWPotential, LogarithmicPotential,\
    CCompositePotential
from ...units import galactic

__all__ = ['LM10Potential'] # ['TriaxialMWPotential']

class LM10Potential(CCompositePotential):
    """
    The Galactic potential used by Law and Majewski (2010) to represent
    the Milky Way as a three-component sum of disk, bulge, and halo.

    The disk potential is an axisymmetric
    :class:`~gala.potential.MiyamotoNagaiPotential`, the bulge potential
    is a spherical :class:`~gala.potential.HernquistPotential`, and the
    halo potential is a triaxial :class:`~gala.potential.LogarithmicPotential`.

    Default parameters are fixed to those found in LM10 by fitting N-body
    simulations to the Sagittarius stream.

    Parameters
    ----------
    units : `~gala.units.UnitSystem` (optional)
        Set of non-reducable units that specify (at minimum) the
        length, mass, time, and angle units.
    disk : dict (optional)
        Parameters to be passed to the :class:`~gala.potential.MiyamotoNagaiPotential`.
    bulge : dict (optional)
        Parameters to be passed to the :class:`~gala.potential.HernquistPotential`.
    halo : dict (optional)
        Parameters to be passed to the :class:`~gala.potential.LogarithmicPotential`.

    Note: in subclassing, order of arguments must match order of potential
    components added at bottom of init.
    """
    def __init__(self, units=galactic,
                 disk=dict(), bulge=dict(), halo=dict()):

        default_disk = dict(m=1E11*u.Msun, a=6.5*u.kpc, b=0.26*u.kpc)
        default_bulge = dict(m=3.4E10*u.Msun, c=0.7*u.kpc)
        default_halo = dict(q1=1.38, q2=1., q3=1.36, r_h=12.*u.kpc,
                            phi=97*u.degree,
                            v_c=np.sqrt(2)*121.858*u.km/u.s)

        for k,v in default_disk.items():
            if k not in disk:
                disk[k] = v

        for k,v in default_bulge.items():
            if k not in bulge:
                bulge[k] = v

        for k,v in default_halo.items():
            if k not in halo:
                halo[k] = v

        super(LM10Potential,self).__init__()

        self["disk"] = MiyamotoNagaiPotential(units=units, **disk)
        self["bulge"] = HernquistPotential(units=units, **bulge)
        self["halo"] = LogarithmicPotential(units=units, **halo)
        self.lock = True

# --------------------------------------------------------------------
# class TriaxialMWPotential(CCompositePotential):

#     def __init__(self, units=galactic,
#                  disk=dict(), bulge=dict(), halo=dict()):
#         """ Axis ratio values taken from Jing & Suto (2002). Other
#             parameters come from a by-eye fit to Bovy's MW2014Potential.
#             Choice of v_c sets circular velocity at Sun to 220 km/s
#         """

#         default_disk = dict(m=7E10, a=3.5, b=0.14)
#         default_bulge = dict(m=1E10, c=1.1)
#         default_halo = dict(a=1., b=0.75, c=0.55,
#                             v_c=0.239225, r_s=30.,
#                             phi=0., theta=0., psi=0.)

#         for k,v in default_disk.items():
#             if k not in disk:
#                 disk[k] = v

#         for k,v in default_bulge.items():
#             if k not in bulge:
#                 bulge[k] = v

#         for k,v in default_halo.items():
#             if k not in halo:
#                 halo[k] = v

#         kwargs = dict()
#         kwargs["disk"] = MiyamotoNagaiPotential(units=units, **disk)
#         kwargs["bulge"] = HernquistPotential(units=units, **bulge)
#         kwargs["halo"] = LeeSutoTriaxialNFWPotential(units=units, **halo)
#         super(TriaxialMWPotential,self).__init__(**kwargs)
# --------------------------------------------------------------------

stuff = """
def busey():
    import webbrowser
    webbrowser.open("http://i.imgur.com/KNoyPwW.jpg")
"""



# coding: utf-8

""" Utilities for Potential classes """

from __future__ import division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# Third-party
import numpy as np

# Project
from .core import PotentialBase

__all__ = ['from_equation']

# def _classnamify(s):
#     s = [x.lower() for x in str(s).split()]
#     words = []
#     for word in s:
#         words.append(word.capitalize())
#     return "".join(words)

def from_equation(expr, vars, pars, name=None, hessian=False):
    r"""
    Create a potential class from an expression for the potential.

    .. note::

        This utility requires having `Sympy <http://www.sympy.org/>`_ installed.

    .. warning::

        These potentials are *not* pickle-able and cannot be written
        out to YAML files (using `~gala.potential.PotentialBase.save()`)

    Parameters
    ----------
    expr : :class:`sympy.core.expr.Expr`, str
        Either a ``Sympy`` expression, or a string that can be converted to
        a ``Sympy`` expression.
    vars : iterable
        An iterable of variable names in the expression.
    pars : iterable
        An iterable of parameter names in the expression.
    name : str (optional)
        The name of the potential class returned.
    hessian : bool (optional)
        Generate a function to compute the Hessian.

    Returns
    -------
    CustomPotential : `~gala.potential.PotentialBase`
        A potential class that represents the input equation. To instantiate the
        potential, use just like a normal class with parameters.

    Examples
    --------
    Here we'll create a potential class for the harmonic oscillator
    potential, :math:`\Phi(x) = \frac{1}{2}\,k\,x^2`::

        >>> Potential = from_equation("1/2*k*x**2", vars="x", pars="k",
        ...                           name='HarmonicOscillator')
        >>> p1 = Potential(k=1.)
        >>> p1
        <HarmonicOscillatorPotential: k=1.00 (dimensionless)>

    The potential class (and object) is a fully-fledged subclass of
    `~gala.potential.PotentialBase` and therefore has many useful methods.
    For example, to integrate an orbit::

        >>> orbit = p1.integrate_orbit([1.,0], dt=0.01, n_steps=1000)

    """
    try:
        import sympy
        from sympy.utilities.lambdify import lambdify
    except ImportError:
        raise ImportError("sympy is required to use 'from_equation()' "
                          "potential class creation.")

    # convert all input to Sympy objects
    expr = sympy.sympify(expr)
    vars = [sympy.sympify(v) for v in vars]
    var_names = [v.name for v in vars]
    pars = [sympy.sympify(p) for p in pars]
    par_names = [p.name for p in pars]

    class CustomPotential(PotentialBase):

        def __init__(self, units=None, **kwargs):
            for par in par_names:
                if par not in kwargs:
                    raise ValueError("You must specify a value for "
                                     "parameter '{}'.".format(par))
            super(CustomPotential,self).__init__(units=units,
                                                 parameters=kwargs)

    if name is not None:
        # name = _classnamify(name)
        if "potential" not in name.lower():
            name = name + "Potential"
        CustomPotential.__name__ = str(name)

    # Energy / value
    valuefunc = lambdify(vars + pars, expr, dummify=False, modules='numpy')
    def _value(self, w, t):
        kw = self.parameters.copy()
        for k,v in kw.items():
            kw[k] = v.value

        for i,name in enumerate(var_names):
            kw[name] = w[i]

        return np.atleast_1d(valuefunc(**kw))
    CustomPotential._value = _value

    # Gradient
    gradfuncs = []
    for var in vars:
        gradfuncs.append(lambdify(vars + pars, sympy.diff(expr,var), dummify=False, modules='numpy'))

    def _gradient(self, w, t):
        kw = self.parameters.copy()
        for k,v in kw.items():
            kw[k] = v.value

        for i,name in enumerate(var_names):
            kw[name] = w[i]

        grad = np.vstack([f(**kw)[np.newaxis] for f in gradfuncs])
        return grad

    CustomPotential._gradient = _gradient
    CustomPotential.save = None

    # Hessian
    if hessian:
        raise NotImplementedError("Hessian functionality doesn't exist yet...sorry.")
        # hessfuncs = []

    return CustomPotential



# coding: utf-8

from __future__ import absolute_import, unicode_literals, division, print_function

__author__ = "adrn <adrn@astro.columbia.edu>"

# This project
from ...units import solarsystem
from ..util import from_equation
from .helpers import PotentialTestBase

class EquationBase(PotentialTestBase):
    def test_plot(self):
        # Skip for now because contour plotting assumes 3D
        pass

    def test_pickle(self):
        # Skip for now because these are not picklable
        pass

    def test_save_load(self):
        # Skip for now because these can't be written to YAML
        pass

class TestHarmonicOscillatorFromEquation(EquationBase):
    Potential = from_equation("1/2*k*x**2", vars="x", pars="k",
                              name='HarmonicOscillator')
    potential = Potential(k=1.)
    w0 = [1.,0.]

class TestHarmonicOscillatorFromEquationUnits(EquationBase):
    Potential = from_equation("1/2*k*x**2", vars="x", pars="k",
                              name='HarmonicOscillator')
    potential = Potential(k=1., units=solarsystem)
    w0 = [1.,0.]

class TestKeplerFromEquation(EquationBase):
    Potential = from_equation("-G*M/sqrt(x**2+y**2+z**2)", vars=["x","y","z"],
                              pars=["G","M"], name='Kepler')
    potential = Potential(G=1., M=1., units=solarsystem)
    w0 = [1.,0.,0.,0.,6.28,0.]
